import base64
import hashlib
import json
import logging
import secrets
import sys
import threading
import urllib
import webbrowser
from time import sleep

import requests
from flask import Flask, render_template, request
from werkzeug.serving import make_server

from suite_py.lib import logger
from suite_py.lib.handler.qainit_handler import QainitHandler as qainit

## Global vars for comunication between flask thread and main cli function
received_callback = None
code = None
error_message = None
received_state = None

## Global Flask App
app = Flask(__name__, template_folder="templates")
log = logging.getLogger("werkzeug")
log.disabled = True


@app.route("/callback")
def callback():
    """
    The callback is invoked after a completed login attempt (succesful or otherwise).
    It sets global variables with the auth code or error messages, then sets the
    polling flag received_callback.
    :return:
    """
    global received_callback, code, error_message, received_state
    error_message = None
    code = None
    if "error" in request.args:
        error_message = request.args["error"] + ": " + request.args["error_description"]
    else:
        code = request.args["code"]
    received_state = request.args["state"]
    received_callback = True
    return render_template("login.html")


class ServerThread(threading.Thread):
    """
    The Flask server is done this way to allow shutting down after a single request has been received.
    """

    def __init__(self, app):
        threading.Thread.__init__(self)
        self.srv = make_server("127.0.0.1", 5000, app)
        self.ctx = app.app_context()
        self.ctx.push()

    def run(self):
        # logger.debug('starting server')
        self.srv.serve_forever()

    def shutdown(self):
        self.srv.shutdown()


class Login:
    def __init__(self, config):

        try:
            if (
                not config.auth0["client_id"]
                or not config.auth0["tenant"]
                or not config.qainit["profile"]
            ):
                self.usage()
                sys.exit(-1)
        except AttributeError:
            self.usage()
            sys.exit(-1)

        self._config = config
        self.auth0_client_id = config.auth0["client_id"]
        self.auth0_tenant = config.auth0["tenant"]
        self.auth0_scope = "openid profile email " + " ".join(
            qainit.scope_mapping[config.qainit["profile"]]
        )
        self.redirect_uri = "http://127.0.0.1:5000/callback"

    def usage(self):
        logger.warning("Unable to login: missing config")
        logger.warning(
            "please check docs: https://github.com/primait/suite_py/blob/master/README.md"
        )

    def auth0_url_encode(self, byte_data):
        """
        Safe encoding handles + and /, and also replace = with nothing
        :param byte_data:
        :return:
        """
        return base64.urlsafe_b64encode(byte_data).decode("utf-8").replace("=", "")

    def generate_challenge(self, a_verifier):
        return self.auth0_url_encode(hashlib.sha256(a_verifier.encode()).digest())

    def run(self):
        global received_callback

        # from https://auth0.com/docs/flows/add-login-using-the-authorization-code-flow-with-pkce
        # Step1: Create code verifier: Generate a code_verifier that will be sent to Auth0 to request tokens.
        verifier = self.auth0_url_encode(secrets.token_bytes(32))
        ## Step2: Create code challenge: Generate a code_challenge from the code_verifier that will be sent to Auth0 to request an authorization_code.
        challenge = self.generate_challenge(verifier)
        state = self.auth0_url_encode(secrets.token_bytes(32))
        client_id = self.auth0_client_id
        tenant = self.auth0_tenant
        scope = self.auth0_scope
        redirect_uri = self.redirect_uri
        #
        # We generate a nonce (state) that is used to protect against attackers invoking the callback
        base_url = f"https://{tenant}.auth0.com/authorize?"
        url_parameters = {
            "audience": "qainit",
            "scope": scope,
            "response_type": "code",
            "redirect_uri": redirect_uri,
            "client_id": client_id,
            "code_challenge": challenge.replace("=", ""),
            "code_challenge_method": "S256",
            "state": state,
        }
        url = base_url + urllib.parse.urlencode(url_parameters)

        # Step3: Authorize user: Request the user's authorization and redirect back to your app with an authorization_code.
        # Open the browser window to the login url
        # Start the server
        # Poll til the callback has been invoked
        received_callback = False
        webbrowser.open(url)
        server = ServerThread(app)
        server.start()
        while not received_callback:
            sleep(1)
        server.shutdown()

        if state != received_state:
            logger.error(
                "Error: session replay or similar attack in progress. Please log out of all connections."
            )
            sys.exit(1)

        if error_message:
            logger.error("An error occurred:")
            logger.error(error_message)
            sys.exit(1)

        # Step4: Request tokens: Exchange your authorization_code and code_verifier for tokens.
        url = f"https://{tenant}.auth0.com/oauth/token"
        headers = {"Content-Type": "application/json"}
        body = {
            "grant_type": "authorization_code",
            "client_id": client_id,
            "code_verifier": verifier,
            "code": code,
            "audience": "https://prima-it-staging.eu.auth0.com/api/v2/",
            "redirect_uri": redirect_uri,
        }
        r = requests.post(url, headers=headers, data=json.dumps(body))
        data = r.json()
        logger.debug(data)
        if "access_token" in data:
            token = str(data["access_token"])
        else:
            logger.error("access_token not found in auth0 response")
            sys.exit(1)
        logger.info("login succeded")
        self._config.put_cache(f"{tenant}_auth0_token", token)



# -*- encoding: utf-8 -*-
import json
import os
import sys

import yaml

from suite_py.lib import logger


class Config:
    def __init__(self, home_path=os.environ["HOME"]):
        self.user = {}
        self.youtrack = {}
        self.vault = {}
        self.aws = {}
        self._config_path_file = os.path.join(home_path, ".suite_py/config.yml")
        self._base_cache_path = os.path.join(home_path, ".suite_py/cache")

        if not os.path.exists(self._base_cache_path):
            os.makedirs(self._base_cache_path)
        self._load()

    def _load(self):
        with open(self._config_path_file, encoding="utf-8") as configfile:
            conf = yaml.safe_load(configfile)

        conf["user"]["projects_home"] = os.path.join(
            os.environ["HOME"], conf["user"]["projects_home"]
        )

        conf["user"].setdefault("review_channel", "#review")
        conf["user"].setdefault("deploy_channel", "#deploy")
        conf["user"].setdefault("default_slug", "PRIMA-XXX")
        conf["user"].setdefault("captainhook_timeout", 30)  # This is in seconds
        conf["user"].setdefault(
            "captainhook_url", "http://captainhook-internal.prima.it"
        )
        conf["user"].setdefault("use_commits_in_pr_body", False)
        conf["user"].setdefault("delete_qa_after_merge", True)

        _load_local_config(conf)

        for k, v in conf.items():
            setattr(self, k, v)

        # AWS section
        try:
            self.aws = conf["aws"]
        except KeyError:
            pass

    def put_cache(self, key, data):
        with open(
            os.path.join(self._base_cache_path, key), "w", encoding="utf-8"
        ) as cache_file:
            json.dump(data, cache_file)

    def get_cache(self, key):
        try:
            with open(
                os.path.join(self._base_cache_path, key), encoding="utf-8"
            ) as cache_file:
                return json.load(cache_file)
        except Exception:
            logger.error(
                f"I couldn't find any cached version for the key {key}. Turn on the VPN."
            )
            sys.exit(-1)


def _load_local_config(conf):
    local_conf_path = os.path.join(os.curdir, ".suite_py.yml")
    try:
        with open(local_conf_path, encoding="utf-8") as f:
            local_conf = yaml.safe_load(f)

            for key in conf.keys():
                conf[key].update(local_conf.get(key, {}))

    except FileNotFoundError:
        pass



import dataclasses
from typing import Dict, List, Optional

import semver
from halo import Halo
from typing_extensions import Literal

from suite_py.lib import logger
from suite_py.lib.handler import git_handler as git
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.github_handler import GithubHandler, GitRelease, Repository

VersionPart = Literal["major", "minor", "patch", "release", "build"]
DEFAULT_VERSION = "0.1.0"


@dataclasses.dataclass
class VersionHandler:
    """
    Manages version-related stuff across different commands.
    """

    _repo: Repository
    _git: git.GitHandler
    _github: GithubHandler

    def select_new_version(
        self,
        current_version: str,
        allow_prerelease: bool = False,
        allow_custom_version: bool = False,
    ) -> str:
        """
        Asks the user for a new semantic version.

        Args:
            current_version: The current project version (for static bumping options)
            allow_prerelease: Whether to give the user the option customize prerelease and build.

        Returns:
            A tuple in the form (selected_part, selected_version).
        """
        bump_options = self.get_bump_options(
            current_version, allow_prerelease, allow_custom_version
        )
        selected_version = prompt_utils.ask_choices("Select version:", bump_options)

        if allow_prerelease and selected_version == "release":
            return self.prompt_prerelease_version(current_version)
        if allow_custom_version and selected_version == "custom":
            return self.prompt_custom_version()
        return selected_version

    def prompt_prerelease_version(self, current_version: str) -> str:
        """
        Asks the user to specify a release token to be used.
        """
        pre_token = prompt_utils.ask_choices(
            "Select prerelease token:", self.get_prerelease_token_options()
        )
        return semver.bump_prerelease(current_version, token=pre_token)

    def prompt_custom_version(self) -> str:
        """
        Asks the user to input a custom semantic version.
        """
        version = prompt_utils.ask_questions_input("Please type a semantic version:")
        while not self.is_valid_semver(version):
            version = prompt_utils.ask_questions_input(
                "Please type a semantic version:"
            )
        return version

    @staticmethod
    def is_valid_semver(version: str) -> bool:
        try:
            semver.VersionInfo.parse(version)
            return True
        except ValueError:
            logger.error(f"The string {version} is not a valid semantic version.")
            return False

    def get_bump_options(
        self, current_version: str, allow_prerelease: bool, allow_custom_version: bool
    ) -> List[Dict[str, str]]:
        """
        Returns a dict of bump alternatives from `current_version`.

        Args:
            current_version: The current project version.

        Returns:
            The mapping between the version part to be bumped and the resulting version.
        """
        options = [
            ("Patch", semver.bump_patch(current_version), None),
            ("Minor", semver.bump_minor(current_version), None),
            ("Major", semver.bump_major(current_version), None),
        ]
        if allow_prerelease:
            options.append(("Release", "release", "(rc, dev, pre...)"))
        if allow_custom_version:
            options.append(("Custom", "custom", "(input a semantic version)"))
        return [
            {"name": f"{name} {desc or value}", "value": value}
            for name, value, desc in options
        ]

    def get_prerelease_token_options(self) -> List[Dict[str, str]]:
        return [{"name": token, "value": token} for token in ["rc", "dev", "pre"]]

    def get_latest_version(self) -> str:
        """
        Retrieves the latest project version using GitHub.

        Returns:
            The latest project version.
        """
        tags = self._repo.get_tags()
        tag = git.get_last_tag_number(tags)
        latest_release = self.get_release(tag)
        current_version = latest_release.tag_name if latest_release else tag
        return current_version or ""

    def get_release(self, tag: str) -> Optional[GitRelease]:
        """
        Retrieves the latest release from GitHub.
        """
        with Halo(text="Loading...", spinner="dots", color="magenta"):
            latest_release = self._github.get_latest_release_if_exists(self._repo)
            if latest_release and latest_release.title == tag:
                return latest_release
        return None



# -*- encoding: utf-8 -*-
import os
import subprocess
import sys
import time
from distutils.version import StrictVersion
from subprocess import CalledProcessError

import requests
import yaml
from halo import Halo

from suite_py.lib import logger

BASEURL = os.environ["DRONE_SERVER"] = "https://drone-1.prima.it"


class DroneHandler:
    def __init__(self, config, tokens, repo=None):
        self._token = tokens.drone
        self._config = config
        os.environ["DRONE_TOKEN"] = self._token
        self._repo = repo
        if repo:
            self._path = os.path.join(config.user["projects_home"], repo)

    def get_last_build_url(self, prefix=None):
        with Halo(text="Contacting drone...", spinner="dots", color="magenta"):
            # necessario per far comparire la build che abbiamo appena pushato
            time.sleep(2)
            try:
                builds = requests.get(
                    f"{BASEURL}/api/repos/primait/{self._repo}/builds",
                    headers={"Authorization": f"Bearer {self._token}"},
                ).json()

                if prefix:
                    builds = [b for b in builds if b["target"].startswith(prefix)]

                return f"{BASEURL}/primait/{self._repo}/{builds[0]['number']}"
            except Exception:
                return ""

    def get_pr_build_number(self, commit_sha):
        with Halo(text="Contacting drone...", spinner="dots", color="magenta"):
            tries = 10
            while tries > 0:
                tries -= 1
                try:
                    builds = requests.get(
                        f"{BASEURL}/api/repos/primait/{self._repo}/builds?per_page=100",
                        headers={"Authorization": f"Bearer {self._token}"},
                    ).json()
                    builds = [b for b in builds if b["after"] == commit_sha]
                    return builds[0]["number"]
                except Exception:
                    time.sleep(1)

            return None

    def get_user(self):
        try:
            user = requests.get(
                f"{BASEURL}/api/user",
                headers={"Authorization": f"Bearer {self._token}"},
            ).json()
            return user
        except Exception:
            return None

    def get_build_url(self, build_number):
        if build_number:
            return f"{BASEURL}/primait/{self._repo}/{build_number}"
        return None

    def get_tags_from_builds(self):
        tags = []
        builds = requests.get(
            f"{BASEURL}/api/repos/primait/{self._repo}/builds?per_page=100",
            headers={"Authorization": f"Bearer {self._token}"},
        ).json()

        for build in builds:
            if build["event"] == "tag":
                tags.append(build["ref"].replace("refs/tags/", ""))

        tags = list(dict.fromkeys(tags))
        tags.sort(key=StrictVersion, reverse=True)
        return tags

    def get_builds_from_tag(self, tag):
        try:
            builds = requests.get(
                f"{BASEURL}/api/repos/primait/{self._repo}/builds?per_page=100&tag={tag}",
                headers={"Authorization": f"Bearer {self._token}"},
            ).json()

            return builds
        except Exception:
            return []

    def get_build_number_from_tag(self, tag):
        attempts = 0
        while attempts < 3:

            try:
                builds = requests.get(
                    f"{BASEURL}/api/repos/primait/{self._repo}/builds?per_page=100&tag={tag}",
                    headers={"Authorization": f"Bearer {self._token}"},
                ).json()

                for build in builds:
                    if build["event"] == "tag":
                        return build["number"]
            except Exception:
                pass

            time.sleep(2)
            attempts += 1

        return None

    def promote(self, build, target, querystring):
        try:
            return requests.post(
                f"{BASEURL}/api/repos/primait/{self._repo}/builds/{build}/promote?target={target}&{querystring}",
                headers={"Authorization": f"Bearer {self._token}"},
            ).json()
        except Exception as e:
            logger.error(f"Unable to trigger drone build, exiting with error: {e}")
            sys.exit(-1)

    def promote_production(self, tag, target, querystring):
        build = self.get_build_number_from_tag(tag)
        if build:
            return self.promote(build, target, querystring)

        logger.error(
            f"Unable to retrieve drone build for repo {self._repo} with tag {tag}"
        )
        sys.exit(-1)

    def launch_build(self, build):
        return requests.post(
            f"{BASEURL}/api/repos/primait/{self._repo}/builds/{build}",
            headers={"Authorization": f"Bearer {self._token}"},
        ).json()

    def prestart_success(self, build_number):
        with Halo(text="Contacting drone...", spinner="dots", color="magenta"):
            tries = 10
            while build_number and tries > 0:
                tries -= 1
                build_status = requests.get(
                    f"{BASEURL}/api/repos/primait/{self._repo}/builds/{build_number}",
                    headers={"Authorization": f"Bearer {self._token}"},
                ).json()

                steps = build_status["stages"][0]["steps"]

                for step in steps:
                    if step["name"] == "pre-start":
                        if step["status"] == "success":
                            return True
                        break
                time.sleep(3)
            return False

    def fmt(self):
        try:
            subprocess.run(
                ["drone", "fmt", "--save", ".drone.yml"], cwd=self._path, check=True
            )
        except (CalledProcessError) as e:
            logger.error(f"{self._repo}: unable to format the .drone.yml {e}")
            sys.exit(-1)

    def validate(self):
        try:
            subprocess.run(
                ["drone", "lint", "--trusted", ".drone.yml"], cwd=self._path, check=True
            )
        except (CalledProcessError) as e:
            logger.error(f"{self._path}: the .drone.yml is not valid {e}")
            sys.exit(-1)

    def sign(self):
        try:
            subprocess.run(
                ["drone", "sign", f"primait/{self._repo}", "--save"],
                cwd=self._path,
                check=True,
            )
        except (CalledProcessError) as e:
            logger.error(f"{self._repo}: unable to sign the .drone.yml {e}")
            sys.exit(-1)

    def parse_yaml(self):
        repo_path = os.path.join(self._config.user["projects_home"], self._repo)
        yaml_path = os.path.join(repo_path, ".drone.yml")
        try:
            return yaml.load_all(
                open(yaml_path, "r", encoding="utf-8"), Loader=yaml.FullLoader
            )
        except FileNotFoundError:
            return None



#################### Maintained by Hatch ####################
# This file is auto-generated by hatch. If you'd like to customize this file
# please add your changes near the bottom marked for 'USER OVERRIDES'.
# EVERYTHING ELSE WILL BE OVERWRITTEN by hatch.
#############################################################
from io import open

from setuptools import find_packages, setup

from suite_py import version


with open("README.rst", "r", encoding="utf-8") as f:
    readme = f.read()

REQUIRES = [
    "PyGithub>=1.51",
    "cement>=3.0",
    "inquirer>=2.7.0",
    "InquirerPy>=0.2.0",
    "PyYaml>=5.4",
    "termcolor>=1.1.0",
    "autoupgrade-prima>=0.6",
    "semver>=2.13.0",
    "halo>=0.0.28",
    "boto3>=1.17.84",
    "Click>=7.0",
    "pptree==3.1",
    "Jinja2>=2.11,<3.0.0",
    "colorama>=0.4.3",
    "Flask==1.1.2",
    "rich==10.1.0",
    "logzero==1.7.0",
    "kubernetes==17.17.0",
    "requests>=2.26.0",
    "requests-toolbelt>=0.9.1",
    "python-dateutil>=2.8.2",
    "itsdangerous==2.0.1",
    "markupsafe==2.0.1",
]

kwargs = {
    "name": "suite-py",
    "version": version,
    "description": "",
    "long_description": readme,
    "author": "Prima Assicurazioni S.p.A",
    "author_email": "devops@prima.it",
    "maintainer": "larrywax, EugenioLaghi, michelangelomo",
    "maintainer_email": "devops@prima.it",
    "url": "https://github.com/primait/suite_py",
    "license": "MIT/Apache-2.0",
    "classifiers": [
        # 'Development Status :: 4 - Beta',
        # 'Intended Audience :: Developers',
        # 'License :: OSI Approved :: MIT License',
        # 'License :: OSI Approved :: Apache Software License',
        # 'Natural Language :: English',
        # 'Operating System :: OS Independent',
        # 'Programming Language :: Python :: 3.5',
        # 'Programming Language :: Python :: 3.6',
        # 'Programming Language :: Python :: Implementation :: CPython',
        # 'Programming Language :: Python :: Implementation :: PyPy',
    ],
    "install_requires": REQUIRES,
    "tests_require": ["coverage", "pytest"],
    "packages": find_packages(exclude=("tests", "tests.*")),
    "package_data": {"templates": ["suite_py/commands/templates/*"]},
    "include_package_data": True,
    "entry_points": {
        "console_scripts": [
            "suite-py = suite_py.cli:main",
        ],
    },
}

#################### BEGIN USER OVERRIDES ####################
# Add your customizations in this section.

###################### END USER OVERRIDES ####################

setup(**kwargs)



# -*- encoding: utf-8 -*-
import json
import sys

import requests
from github.GithubException import UnknownObjectException
from halo import Halo

from suite_py.lib import logger
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.git_handler import GitHandler
from suite_py.lib.handler.github_handler import GithubHandler


class QainitHandler:

    scope_mapping = {
        "admin": [
            "update:user-quota",
            "delete:others-qa",
            "create:qa",
            "delete:qa",
            "describe:qa",
            "describe:others-qa",
            "update:qa",
            "update:others-qa",
            "list:qa",
            "list:others-qa",
        ],
        "dev": [
            "create:qa",
            "delete:qa",
            "describe:qa",
            "describe:others-qa",
            "update:qa",
            "update:others-qa",
            "list:qa",
            "list:others-qa",
        ],
        "external": ["create:qa", "delete:qa", "describe:qa", "list:qa", "update:qa"],
    }

    def __init__(self, project, config, tokens):
        self._project = project
        self._token = tokens.drone
        self._config = config
        self.url = self._config.qainit["url"]
        self.auth0_tenant = self._config.auth0["tenant"]
        self.auth0_token = self._config.get_cache(f"{self.auth0_tenant}_auth0_token")
        self._github = GithubHandler(tokens)
        self._git = GitHandler(project, config)

        if "url" not in config.qainit:
            self.usage()
            sys.exit(-1)

    def usage(self):
        logger.warning(
            "Unable to use QA commands: missing qainit config in ~/.suite_py/config.yml"
        )
        logger.warning(
            "Update your config.yml as: https://github.com/primait/suite_py/blob/master/.config.yml.dist"
        )

    def user_info(self):
        r = self._execute("GET", "/api/v1/user").json()
        logger.debug(json.dumps(r, indent=2))

        return r

    def update_user_quota(self, username, quota):
        body = {"github_username": f"{username}", "quota": f"{quota}"}
        logger.debug(json.dumps(body))
        r = self._execute(
            "POST",
            "/api/v1/user/quota",
            body=json.dumps(body),
        )

        logger.info("Quota updated.")
        logger.debug(json.dumps(r.json(), indent=2))

    def create(self, name, card, services):
        srv_list = self.create_services_body(services)
        body = {"name": name, "card": card, "services": srv_list}

        logger.debug(json.dumps(body))
        r = self._execute(
            "POST",
            "/api/v1/qa",
            body=json.dumps(body),
        )

        logger.info(f"QA creation initiated. Your namespace hash: {r.json()['hash']}")
        logger.debug(json.dumps(r.json(), indent=2))

    def update(self, qa_hash, services):
        srv_list = self.create_services_body(services)

        body = {"services": srv_list}
        logger.debug(json.dumps(body))

        r = self._execute(
            "PUT",
            f"/api/v1/qa/{qa_hash}",
            body=json.dumps(body),
        )
        logger.info("QA update initiated")
        logger.debug(json.dumps(r.json(), indent=2))

    def list(self, params, page=1, page_size=10):
        filters = []
        status_values = '"created","creating","updated","updating","failed","frozen","freezing","unfreezing"'
        if len(params["status"]) > 0:
            status_values = ",".join([f'"{status}"' for status in params["status"]])
        filters.append(f"status=[{status_values}]")

        if params["user"] is not None:
            filters.append(f"user={params['user']}")

        if params["card"] is not None:
            filters.append(f"card={params['card']}")

        filters_string = "&".join(filters)

        return self._execute(
            "GET",
            f"/api/v1/qa?{filters_string}&page_size={page_size}&page={page}",
        )

    def delete(self, qa_hash):
        r = self._execute(
            "DELETE",
            f"/api/v1/qa/{qa_hash}",
        )
        logger.info("QA deletion initiated")
        logger.debug(json.dumps(r.json(), indent=2))

    def describe(self, qa_hash):
        r = self._execute(
            "GET",
            f"/api/v1/qa/{qa_hash}",
        ).json()
        logger.debug(json.dumps(r, indent=2))

        return r

    def create_services_body(self, prj_list):
        srv_list = []
        ref = self._git.current_branch_name()
        for prj in prj_list:
            with Halo(text="Loading branches...", spinner="dots", color="magenta"):
                choices = [
                    {"name": branch.name, "value": branch.name}
                    for branch in self._github.get_branches(prj)
                ]
            if choices:
                choices.sort(key=lambda x: x["name"])
                ref = prompt_utils.ask_choices(
                    f"Select branch for project - {prj}: ", choices, default_text=ref
                )
            try:
                self._github.get_raw_content(prj, ref, ".service.yml")
            except UnknownObjectException:
                logger.error(
                    f".service.yml missing for project: {prj}, can't add microservice to QA"
                )
                sys.exit(-1)
            srv_list.append(
                {
                    "name": prj,
                    "ref": ref,
                }
            )

        return srv_list

    def freeze(self, qa_hash):
        body = {"operation": "freeze"}
        logger.debug(json.dumps(body))
        r = self._execute("PUT", f"/api/v1/qa/{qa_hash}", body=json.dumps(body))
        logger.info("QA freezing initiated")
        logger.debug(json.dumps(r.json(), indent=2))

    def unfreeze(self, qa_hash):
        body = {"operation": "unfreeze"}
        logger.debug(json.dumps(body))
        r = self._execute("PUT", f"/api/v1/qa/{qa_hash}", body=json.dumps(body))
        logger.info("QA unfreezing initiated")
        logger.debug(json.dumps(r.json(), indent=2))

    def maintenance(self):
        r = self._execute("POST", "/api/v1/maintenance")
        if r.json()["success"]:
            maintenance = r.json()["maintenance"]
            logger.info(
                f"Maintenance mode is now {'enabled' if maintenance else 'disabled'}"
            )

        logger.debug(json.dumps(r.json(), indent=2))

    def _execute(self, request_method, api_endpoint, body=None):
        api_url = self.url + api_endpoint
        auth0_token = self._config.get_cache(f"{self.auth0_tenant}_auth0_token")

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + auth0_token,
        }
        logger.debug(request_method)
        logger.debug(api_url)
        logger.debug(headers)
        logger.debug(body)
        r = requests.request(request_method, api_url, headers=headers, data=body)

        if 200 <= r.status_code <= 299:
            logger.debug("Call to qainit-evo executed successfully")
        else:
            logger.error("Some issue during call to qainit-evo: ")
            logger.error(f"Status code: {r.status_code}, response: {r.text}")
            logger.error(api_endpoint)
            sys.exit(-1)

        return r



# -*- encoding: utf-8 -*-
import sys

import requests

from suite_py.lib import logger
from suite_py.lib.handler import git_handler as git


class CaptainHook:
    def __init__(self, config):
        self._baseurl = config.user["captainhook_url"]
        self._timeout = config.user["captainhook_timeout"]

    def lock_project(self, project, env):
        data = {
            "project": project,
            "status": "locked",
            "user": git.get_username(),
            "environment": env,
        }
        return self.send_post_request("/projects/manage-lock", data)

    def unlock_project(self, project, env):
        data = {
            "project": project,
            "status": "unlocked",
            "user": git.get_username(),
            "environment": env,
        }
        return self.send_post_request("/projects/manage-lock", data)

    def status(self, project, env):
        return self.send_get_request(
            f"/projects/check?project={project}&environment={env}"
        )

    def check(self):
        return requests.get(f"{self._baseurl}/", timeout=(2, self._timeout))

    def get_users_list(self):
        return self.send_get_request("/users/all")

    def aggregators(self):
        return self.send_get_request("/cloudflare/aggregators/available")

    def change_aggregator(self, aggregator, qa_address):
        data = {"aggregator": aggregator, "qa_address": qa_address}
        return self.send_put_request("/cloudflare/aggregators", data)

    def send_post_request(self, endpoint, data):
        try:
            return requests.post(
                f"{self._baseurl}{endpoint}", data=data, timeout=self._timeout
            )
        except Exception:
            logger.error("Unable to contact Captainhook, are you using the VPN?")
            sys.exit(-1)

    def send_put_request(self, endpoint, data):
        try:
            return requests.put(
                f"{self._baseurl}{endpoint}", data=data, timeout=self._timeout
            )
        except Exception:
            logger.error("Unable to contact Captainhook, are you using the VPN?")
            sys.exit(-1)

    def send_get_request(self, endpoint):
        try:
            return requests.get(
                f"{self._baseurl}{endpoint}", timeout=(2, self._timeout)
            )
        except Exception:
            logger.error("Unable to contact Captainhook, are you using the VPN?")
            sys.exit(-1)

    def set_timeout(self, timeout):
        self._timeout = timeout



# -*- coding: utf-8 -*-
import os
import re

from pptree import Node, print_tree

from suite_py.lib import logger
from suite_py.lib.handler.aws_handler import Aws


class IP:
    def __init__(self, project, config, env):
        self._project = project
        self._env = env
        self._aws = Aws(config)

    def run(self):

        clusters_names = self._aws.get_ecs_clusters(self._env)
        n_services = Node("services")

        projects = {"prima": ["web", "consumer-api"], "ab_normal": ["abnormal"]}
        project_names = projects.get(self._project, [self._project])

        for cluster_name in clusters_names:

            services = []
            all_services = self._aws.get_ecs_services(cluster_name)

            for service in all_services:
                if service["status"] == "ACTIVE":
                    for prj in project_names:
                        if prj in service["serviceName"]:
                            services.append(service["serviceName"])

            for service in services:
                container_instances = []
                container_instances = (
                    self._aws.get_container_instances_arn_from_service(
                        cluster_name, service
                    )
                )
                if container_instances:
                    ips = self._aws.get_ips_from_container_instances(
                        cluster_name, container_instances
                    )

                    m = re.search(f"ecs-task-.*-{self._env}-ECSService(.*)-.*", service)
                    if m:
                        n_service = Node(m.group(1), n_services)
                        for ip in ips:
                            Node(ip, n_service)

        if self._env == "staging" and os.path.isfile("./deploy/values/staging.yml"):
            logger.info(
                "This project has been migrated on k8s, command is no longer available."
            )
            logger.info(
                "Guide on how to setup k8s: https://app.tettra.co/teams/prima-assicurazioni/pages/effettuare-operazioni-su-k8s"
            )
        elif n_services.children:
            print_tree(n_services, horizontal=True)
        else:
            logger.info(
                f"No active tasks for {self._project} in environment {self._env}"
            )
        logger.info("Done!")



# -*- coding: utf-8 -*-
import functools
import readline
import sys

from suite_py.lib import logger
from suite_py.lib.handler.captainhook_handler import CaptainHook
from suite_py.lib.handler.git_handler import GitHandler
from suite_py.lib.handler.github_handler import GithubHandler
from suite_py.lib.handler.youtrack_handler import YoutrackHandler


class AskReview:
    def __init__(self, project, config, tokens):
        self._project = project
        self._config = config
        self._youtrack = YoutrackHandler(config, tokens)
        self._captainhook = CaptainHook(config)
        self._git = GitHandler(project, config)
        self._github = GithubHandler(tokens)

    def run(self):
        users = self._maybe_get_users_list()
        pr = self._get_pr()
        youtrack_reviewers = _ask_reviewer(users)
        github_reviewers = _find_github_nicks(youtrack_reviewers, users)
        pr.create_review_request(github_reviewers)
        logger.info("Adding reviewers on GitHub")
        self._maybe_adjust_youtrack_card(pr.title, youtrack_reviewers)

    def _maybe_get_users_list(self):
        try:
            users = self._captainhook.get_users_list().json()
            self._config.put_cache("users", users)
            return users
        except Exception:
            logger.warning(
                "Can't get users list from Captainhook. Using cached version."
            )
            return self._config.get_cache("users")

    def _get_pr(self):
        branch_name = self._git.current_branch_name()
        pull = self._github.get_pr_from_branch(self._project, branch_name)

        if pull.totalCount:
            pr = pull[0]
            logger.info(
                f"I found pull request number {pr.number} for branch {branch_name} on repo {self._project}"
            )
        else:
            logger.error(f"No open pull requests found for branch {branch_name}")
            sys.exit(-1)

        return pr

    def _maybe_adjust_youtrack_card(self, title, youtrack_reviewers):
        youtrack_id = self._youtrack.get_card_from_name(title)

        if youtrack_id:
            logger.info(
                f"Moving the {youtrack_id} card for review on youtrack and adding user tags"
            )
            self._youtrack.update_state(
                youtrack_id, self._config.youtrack["review_state"]
            )
            for rev in youtrack_reviewers:
                try:
                    self._youtrack.add_tag(youtrack_id, f"review:{rev}")
                except BaseException as e:
                    logger.warning(f"I was unable to add the review tags: {e}")
                    sys.exit(-1)
        else:
            logger.warning(
                "Reviewers added ONLY on GitHub. No linked card on YouTrack or missing card number in the branch name."
            )


def _ask_reviewer(users):
    u_completer = functools.partial(_completer, users)
    readline.set_completer(u_completer)
    readline.parse_and_bind("tab: complete")

    youtrack_reviewers = []

    youtrack_reviewers = list(
        input(
            "Choose the reviewers (name.surname - separated by space - press TAB for autocomplete) > "
        ).split()
    )

    if not youtrack_reviewers:
        logger.warning("You must enter at least one reviewer")
        return _ask_reviewer(users)

    return youtrack_reviewers


def _completer(users, text, state):
    options = [x["youtrack"] for x in users if text.lower() in x["youtrack"].lower()]
    try:
        return options[state]
    except IndexError:
        return None


def _find_github_nicks(youtrack_reviewers, users):
    github_reviewers = []
    for rev in youtrack_reviewers:
        for user in users:
            if user["youtrack"] == rev:
                github_reviewers.append(user["github"])

    return github_reviewers



# -*- coding: utf-8 -*-
import sys

from halo import Halo

from suite_py.lib import logger
from suite_py.lib.handler import git_handler as git
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.captainhook_handler import CaptainHook
from suite_py.lib.handler.drone_handler import DroneHandler
from suite_py.lib.handler.git_handler import GitHandler
from suite_py.lib.handler.github_handler import GithubHandler
from suite_py.lib.handler.qainit_handler import QainitHandler
from suite_py.lib.handler.youtrack_handler import YoutrackHandler
from suite_py.lib.symbol import CHECKMARK, CROSSMARK


class MergePR:
    def __init__(self, project, config, tokens):
        self._project = project
        self._config = config
        self._youtrack = YoutrackHandler(config, tokens)
        self._captainhook = CaptainHook(config)
        self._git = GitHandler(project, config)
        self._github = GithubHandler(tokens)
        self._drone = DroneHandler(config, tokens, repo=project)
        self._qainit = QainitHandler(project, config, tokens)

    def run(self):
        self._stop_if_master_locked()

        pr = self._select_pr()

        print(f"\nYou have selected: \n{self._check_pr_status(pr)}")
        if not prompt_utils.ask_confirm(
            "Do you want to continue the merge?", default=False
        ):
            sys.exit()

        branch_name = pr.head.ref
        youtrack_id = self._youtrack.get_card_from_name(branch_name)

        _check_migrations_merge(pr)

        logger.info("Merging...")

        merge_status = pr.merge(
            commit_title=f"{pr.title} (#{pr.number})",
            commit_message="",
            merge_method="squash",
        )

        if not merge_status.merged:
            logger.error("An error occurred during the merge.")
            sys.exit(-1)

        drone_build_number = self._drone.get_pr_build_number(merge_status.sha)
        drone_build_url = self._drone.get_build_url(drone_build_number)

        if drone_build_url:
            logger.info(
                f"Pull request merged on master! You can follow the build status at {drone_build_url}"
            )
        else:
            logger.info("Pull request merged on master!")

        self._git.fetch()
        if self._git.remote_branch_exists(branch_name):
            self._git.delete_remote_branch(branch_name)

        if prompt_utils.ask_confirm(
            "Do you want to lock staging? (If you need to test on staging)",
            default=False,
        ):
            self._captainhook.lock_project(self._project, "staging")

        if youtrack_id:
            logger.info("Updating card status on YouTrack...")
            self._youtrack.update_state(
                youtrack_id, self._config.youtrack["merged_state"]
            )
            logger.info("Card updated")

            if self._config.user["delete_qa_after_merge"]:
                with Halo(
                    text="Searching for related QA environments...",
                    spinner="dots",
                    color="magenta",
                ):
                    response = self._qainit.list(
                        {
                            "user": "all",
                            "card": youtrack_id,
                            "status": ["created", "updated", "failed", "frozen"],
                        }
                    ).json()

                for qa in response["list"]:
                    logger.info(f"Found {qa['name']}")

                if len(response["list"]) > 0 and prompt_utils.ask_confirm(
                    "Do you want to delete the QA env(s) I found?",
                    default=True,
                ):
                    for qa in response["list"]:
                        logger.info(f"Deleting {qa['name']} ")
                        self._qainit.delete(qa["hash"])

        else:
            logger.warning(
                "There is no YouTrack issue in the branch name or the selected issue does not exist."
            )
            logger.warning(
                "No card updated on YouTrack and no QA turned off automatically"
            )

        logger.info("All done!")
        sys.exit()

    def _select_pr(self):
        if self._github.user_is_admin(self._project):
            logger.warning(
                "You are admin of the repository, you can merge it skipping the checks (CI, review, etc ...)\nWith great power comes great responsibility"
            )

        with Halo(text="Loading pull requests...", spinner="dots", color="magenta"):
            choices = [
                {"name": pr.title, "value": pr}
                for pr in self._github.get_list_pr(self._project)
            ]
        if choices:
            choices.sort(key=lambda x: x["name"])
            return prompt_utils.ask_choices("Select PR: ", choices)

        logger.error(
            f"There are no pull requests ready to be merged or you may not have the permissions, please check on https://github.com/primait/{self._project}/pulls"
        )
        sys.exit(-1)

    def _stop_if_master_locked(self):
        request = self._captainhook.status(self._project, "staging")

        if request.status_code != 200:
            logger.error("Unable to determine lock status on master.")
            sys.exit(-1)

        request_object = request.json()
        if request_object["locked"]:
            logger.error(
                f"The project is locked on staging by {request_object ['by']}. Unable to continue."
            )
            sys.exit(-1)

    def _check_pr_status(self, pr):
        build_status = CHECKMARK
        reviews = CHECKMARK
        print(pr.mergeable_state)
        if pr.mergeable_state == "dirty":
            logger.error(
                "The selected pull request is not mergeable. Check for conflicts."
            )
            sys.exit(-1)
        if pr.mergeable_state == "blocked":
            build_status = self._pr_build_status(pr)
            reviews = _pr_reviews(pr)

        return f"#{pr.number} {pr.title}\n     build: {build_status} - reviews: {reviews}\n"

    def _pr_build_status(self, pr):
        if self._github.get_build_status(self._project, pr.head.sha).state == "success":
            return CHECKMARK
        return CROSSMARK


def _check_migrations_merge(pr):
    files_changed = [x.filename for x in pr.get_files()]
    if git.migrations_found(files_changed):
        logger.warning("WARNING: migrations detected in the code")
        if not prompt_utils.ask_confirm("Are you sure you want to continue?"):
            sys.exit()


def _pr_reviews(pr):
    reviews = [r for r in pr.get_reviews() if r.state == "APPROVED"]
    if reviews:
        return CHECKMARK
    return CROSSMARK
